from guniflask.app.initializer import AppInitializer
from guniflask.config import load_app_env


def create_app(with_context=True):
    load_app_env()
    app_initializer = AppInitializer()
    app = app_initializer.init(with_context=with_context)
    return app



from guniflask.context import configuration
from guniflask.security_config import WebSecurityConfigurer, HttpSecurity


@configuration
class SecurityConfiguration(WebSecurityConfigurer):

    def configure_http(self, http: HttpSecurity):
        """Configure http security here"""



from typing import Any


class FieldInfo:
    __slots__ = (
        'name',
        'default',
        'dtype',
        'description',
        'required',
        'extra',
    )

    def __init__(self, **kwargs):
        self.name = kwargs.pop('name', None)
        self.default = kwargs.pop('default', None)
        self.dtype = kwargs.pop('dtype', None)
        self.description = kwargs.pop('description', None)
        self.required = kwargs.pop('required', None)
        self.extra = kwargs


def RequestParam(name: str = None,
                 default: Any = None,
                 dtype: type = None,
                 description: str = None,
                 required: bool = None,
                 **extra: Any) -> Any:
    return RequestParamInfo(name=name,
                            default=default,
                            dtype=dtype,
                            description=description,
                            required=required,
                            **extra)


class RequestParamInfo(FieldInfo):
    pass


def PathVariable(name: str = None,
                 default: Any = None,
                 dtype: type = None,
                 description: str = None,
                 required: bool = None,
                 **extra: Any) -> Any:
    return PathVariableInfo(name=name,
                            default=default,
                            dtype=dtype,
                            description=description,
                            required=required,
                            **extra)


class PathVariableInfo(FieldInfo):
    pass


def RequestBody(name: str = None,
                default: Any = None,
                dtype: type = None,
                description: str = None,
                required: bool = None,
                **extra: Any) -> Any:
    return RequestBodyInfo(name=name,
                           default=default,
                           dtype=dtype,
                           description=description,
                           required=required,
                           **extra)


class RequestBodyInfo(FieldInfo):
    pass


def RequestHeader(name: str = None,
                  default: Any = None,
                  dtype: type = None,
                  description: str = None,
                  required: bool = None,
                  **extra: Any) -> Any:
    return RequestHeaderInfo(name=name,
                             default=default,
                             dtype=dtype,
                             description=description,
                             required=required,
                             **extra)


class RequestHeaderInfo(FieldInfo):
    pass


def CookieValue(name: str = None,
                default: Any = None,
                dtype: type = None,
                description: str = None,
                required: bool = None,
                **extra: Any) -> Any:
    return CookieValueInfo(name=name,
                           default=default,
                           dtype=dtype,
                           description=description,
                           required=required,
                           **extra)


class CookieValueInfo(FieldInfo):
    pass


def FormValue(name: str = None,
              default: Any = None,
              dtype: type = None,
              description: str = None,
              required: bool = None,
              **extra: Any) -> Any:
    return FormValueInfo(name=name,
                         default=default,
                         dtype=dtype,
                         description=description,
                         required=required,
                         **extra)


class FormValueInfo(FieldInfo):
    pass


def FilePart(name: str = None,
             default: Any = None,
             dtype: type = None,
             description: str = None,
             required: bool = None,
             **extra: Any) -> Any:
    return FilePartInfo(name=name,
                        default=default,
                        dtype=dtype,
                        description=description,
                        required=required,
                        **extra)


class FilePartInfo(FieldInfo):
    pass



from abc import ABCMeta, abstractmethod

from guniflask.context.annotation import configuration, bean
from guniflask.scheduling.async_executor import AsyncExecutor, DefaultAsyncExecutor
from guniflask.scheduling.async_post_processor import AsyncPostProcessor
from guniflask.scheduling.config_constants import *


class AsyncConfigurer(metaclass=ABCMeta):
    @abstractmethod
    def get_async_executor(self) -> AsyncExecutor:
        pass  # pragma: no cover


@configuration
class AsyncConfiguration:
    def __init__(self, async_configurer: AsyncConfigurer = None):
        if async_configurer is None:
            self._async_executor = DefaultAsyncExecutor()
        else:
            self._async_executor = async_configurer.get_async_executor()

    @bean(ASYNC_ANNOTATION_PROCESSOR)
    def async_annotation_processor(self) -> AsyncPostProcessor:
        return AsyncPostProcessor()

    @bean
    def async_executor(self) -> AsyncExecutor:
        return self._async_executor



from abc import ABCMeta, abstractmethod
from typing import Any


class SecurityBuilder(metaclass=ABCMeta):
    @abstractmethod
    def build(self):
        pass  # pragma: no cover


class AbstractSecurityBuilder(SecurityBuilder, metaclass=ABCMeta):
    def __init__(self):
        self._object: Any = None

    def build(self) -> Any:
        self._object = self._do_build()
        return self._object

    @property
    def object(self) -> Any:
        return self._object

    @abstractmethod
    def _do_build(self) -> Any:
        pass  # pragma: no cover



from .annotation import autowired
from .annotation import bean
from .annotation import component
from .annotation import condition_on_setting
from .annotation import conditional
from .annotation import configuration
from .annotation import controller
from .annotation import include
from .annotation import repository
from .annotation import service
from .annotation_config_registry import AnnotatedBeanDefinitionReader
from .annotation_config_registry import AnnotationConfigRegistry
from .annotation_config_registry import ModuleBeanDefinitionScanner
from .annotation_config_utils import AnnotationConfigUtils
from .bean_context import BeanContext
from .bean_context import BeanContextAware
from .default_bean_context import AnnotationConfigBeanContext
from .default_bean_context import DefaultBeanContext
from .event import ApplicationEvent
from .event import ContextClosedEvent
from .event import ContextRefreshedEvent
from .event_listener import ApplicationEventListener
from .event_publisher import ApplicationEventPublisher



from flask import current_app

from guniflask.config.app_settings import settings
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.authentication_provider import AuthenticationProvider
from guniflask.security_config.authentication_manager_builder import AuthenticationManagerBuilder
from guniflask.security_config.cors_configurer import CorsConfigurer
from guniflask.security_config.http_basic_configurer import HttpBasicConfigurer
from guniflask.security_config.http_security_builder import HttpSecurityBuilder
from guniflask.web.gzip import GzipFilter
from guniflask.web.request_filter import RequestFilter, RequestFilterChain


class HttpSecurity(HttpSecurityBuilder):
    def __init__(self, authentication_builder: AuthenticationManagerBuilder,
                 shared_objects: dict = None):
        super().__init__()
        assert authentication_builder is not None, 'Authentication builder is required'
        self.set_shared_object(AuthenticationManagerBuilder, authentication_builder)
        if shared_objects:
            for k, v in shared_objects.items():
                self.set_shared_object(k, v)
        self._security_filter_chain = RequestFilterChain()
        self._blueprints = []

    def _perform_build(self):
        app = current_app._get_current_object()
        app.before_request(self._security_filter_chain.before_request)
        app.after_request(self._security_filter_chain.after_request)

    def _before_configure(self):
        cors_config = settings.get_by_prefix('guniflask.cors')
        if cors_config is not None:
            self.cors(cors_config)

        gzip_config = settings.get_by_prefix('guniflask.gzip_on')
        if gzip_config:
            if isinstance(gzip_config, dict):
                self.add_request_filter(GzipFilter(**gzip_config))
            else:
                self.add_request_filter(GzipFilter())

        self.set_shared_object(AuthenticationManager, self._get_authentication_registry().build())

    def with_authentication_provider(self, authentication_provider: AuthenticationProvider) -> 'HttpSecurity':
        self._get_authentication_registry().with_authentication_provider(authentication_provider)
        return self

    def _get_authentication_registry(self) -> AuthenticationManagerBuilder:
        return self.get_shared_object(AuthenticationManagerBuilder)

    def http_basic(self) -> HttpBasicConfigurer:
        return self._get_or_apply(HttpBasicConfigurer())

    def cors(self, config) -> CorsConfigurer:
        return self._get_or_apply(CorsConfigurer(config))

    def _get_or_apply(self, configurer):
        existing = self.get_configurer(type(configurer))
        if existing:
            return existing
        return self.apply(configurer)

    def add_request_filter(self, request_filter: RequestFilter) -> 'HttpSecurity':
        self._security_filter_chain.add_request_filter(request_filter)
        return self



from typing import List

from guniflask.security_config.configured_security_builder import ConfiguredSecurityBuilder
from guniflask.security_config.security_builder import SecurityBuilder


class WebSecurity(ConfiguredSecurityBuilder):
    def __init__(self):
        super().__init__()
        self._security_builders: List[SecurityBuilder] = []

    def _perform_build(self):
        for builder in self._security_builders:
            builder.build()

    def add_security_builder(self, security_builder: SecurityBuilder):
        self._security_builders.append(security_builder)



from abc import ABCMeta, abstractmethod

from guniflask.context.event import ApplicationEvent


class ApplicationEventListener(metaclass=ABCMeta):

    @abstractmethod
    def on_application_event(self, application_event: ApplicationEvent):
        pass  # pragma: no cover



import inspect
from importlib import import_module

from guniflask.annotation import AnnotationMetadata, AnnotationUtils
from guniflask.beans.definition import BeanDefinition
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.beans.factory_post_processor import BeanDefinitionRegistryPostProcessor
from guniflask.beans.name_generator import BeanNameGenerator
from guniflask.beans.singleton_registry import SingletonBeanRegistry
from guniflask.context.annotation import Bean, Component, Configuration, Include
from guniflask.context.bean_name_generator import AnnotationBeanNameGenerator
from guniflask.context.condition_evaluator import ConditionEvaluator
from guniflask.context.config_constants import *


class ConfigurationClassPostProcessor(BeanDefinitionRegistryPostProcessor):

    def __init__(self):
        self._component_bean_name_generator = AnnotationBeanNameGenerator()

    def post_process_bean_definition_registry(self, registry: BeanDefinitionRegistry):
        self._process_config_bean_definitions(registry)

    def _process_config_bean_definitions(self, registry: BeanDefinitionRegistry):
        if isinstance(registry, SingletonBeanRegistry):
            singleton_registry = registry
            if singleton_registry.contains_singleton(CONFIGURATION_BEAN_NAME_GENERATOR):
                generator = singleton_registry.get_singleton(CONFIGURATION_BEAN_NAME_GENERATOR)
                self._component_bean_name_generator = generator

        candidate_names = registry.get_bean_definition_names()
        candidates = {}
        for name in candidate_names:
            bean_definition = registry.get_bean_definition(name)
            if self._is_configuration_class_candidate(bean_definition):
                candidates[name] = bean_definition

        reader = ConfigurationClassBeanDefinitionReader(registry, self._component_bean_name_generator)
        while len(candidates) > 0:
            for bean_name, bean_definition in candidates.items():
                reader.load_bean_definitions(bean_name, bean_definition)

            candidates = {}
            new_candidate_names = registry.get_bean_definition_names()
            if len(new_candidate_names) > len(candidate_names):
                old_candidate_names = set(candidate_names)
                for name in new_candidate_names:
                    if name not in old_candidate_names:
                        bean_definition = registry.get_bean_definition(name)
                        if self._is_configuration_class_candidate(bean_definition):
                            candidates[name] = bean_definition
                candidate_names = new_candidate_names

    def _is_configuration_class_candidate(self, bean_definition: BeanDefinition):
        annotation_metadata = AnnotationUtils.get_annotation_metadata(bean_definition.source)
        if annotation_metadata is None:
            return False
        if annotation_metadata.is_annotated(Configuration):
            return True
        if annotation_metadata.is_annotated(Component):
            return True
        if annotation_metadata.is_annotated(Include):
            return True
        return False


class ConfigurationClassBeanDefinitionReader:
    def __init__(self, registry: BeanDefinitionRegistry, include_bean_name_generator: BeanNameGenerator):
        self._registry = registry
        self._condition_evaluator = ConditionEvaluator(registry)
        self._include_bean_name_generator = include_bean_name_generator

    def load_bean_definitions(self, bean_name: str, bean_definition: BeanDefinition):
        source = bean_definition.source
        meta_data = AnnotationUtils.get_annotation_metadata(source)
        if meta_data.is_annotated(Include):
            self._load_bean_definition_for_included_config(meta_data)

        for m in dir(source):
            func = getattr(source, m)
            if inspect.isfunction(func):
                method_metadata = AnnotationUtils.get_annotation_metadata(func)
                if method_metadata is not None and method_metadata.is_annotated(Bean):
                    self._load_bean_definition_for_bean_method(bean_name, method_metadata)

    def _load_bean_definition_for_bean_method(self, factory_bean_name: str, method_metadata: AnnotationMetadata):
        if self._condition_evaluator.should_skip(method_metadata):
            return
        method = method_metadata.source
        method_name = method.__name__
        bean_definition = BeanDefinition(method)
        bean_definition.factory_bean_name = factory_bean_name
        attributes = method_metadata.get_annotation(Bean).attributes
        bean_name = attributes.get('name') or method_name
        self._registry.register_bean_definition(bean_name, bean_definition)

    def _load_bean_definition_for_included_config(self, metadata: AnnotationMetadata):
        annotation = metadata.get_annotation(Include)
        config_set = annotation['values']
        if config_set:
            for config_cls in config_set:
                if isinstance(config_cls, str):
                    _m, _c = config_cls.rsplit('.', maxsplit=1)
                    config_cls = getattr(import_module(_m), _c)
                config_cls_metadata = AnnotationUtils.get_annotation_metadata(config_cls)
                if self._condition_evaluator.should_skip(config_cls_metadata):
                    continue
                bean_definition = BeanDefinition(config_cls)
                bean_name = self._include_bean_name_generator.generate_bean_name(bean_definition, self._registry)
                self._registry.register_bean_definition(bean_name, bean_definition)



from flask import jsonify

from guniflask.web import blueprint, post_route, RequestBody, get_route, FormValue, FilePart, CookieValue, RequestHeader


@blueprint
class ParamController:
    @get_route('/get/int')
    def get_int(self, x: int):
        return {'x': x}

    @get_route('/get/str')
    def get_str(self, x: str):
        return {'x': x}

    @get_route('/get/default')
    def get_default(self, x: str = 'default'):
        return {'x': x}

    @post_route('/post/dict')
    def post_dict(self, data: dict = RequestBody):
        assert isinstance(data, dict), type(data)
        return data

    @post_route('/post/list')
    def post_list(self, data: list = RequestBody):
        assert isinstance(data, list), type(data)
        return jsonify(data)

    @post_route('/post/form-int')
    def post_form_int(self, x: int = FormValue):
        return {'x': x}

    @post_route('/post/form-str')
    def post_form_str(self, x: str = FormValue):
        return {'x': x}

    @post_route('/post/file-bytes')
    def post_file_bytes(self, x: bytes = FilePart):
        assert isinstance(x, bytes)
        return {'x': x.decode()}

    @post_route('/post/cookie-str')
    def post_cookie_str(self, x: str = CookieValue):
        return {'x': x}

    @post_route('/post/cookie-int')
    def post_cookie_int(self, x: int = CookieValue):
        return {'x': x}

    @post_route('/post/header-int')
    def post_header_int(self, x: int = RequestHeader):
        return {'x': x}

    @post_route('/post/header-str')
    def post_header_str(self, x: str = RequestHeader):
        return {'x': x}



from guniflask.web import blueprint, get_route


@blueprint
class GzipController:
    @get_route('/gzip/json-512')
    def get_json_512(self):
        return {
            'result': 'a' * 512
        }

    @get_route('/gzip/json-1024')
    def get_json_1024(self):
        return {
            'result': 'a' * 1024
        }
